#include "match.h"
#include "api.h"
#include "utils.h"

#include <unordered_set>
#include <iostream>

// ****************************************************************************
// Utilities so that we can use Row as the key in a std::unordered_set
// ****************************************************************************

/* Calculates a hash value of a Row */
template<>
struct std::hash<Row> {
    size_t operator()(const Row& row) const {
        size_t hash = 0;
        for(size_t c = 0; c < row.numCols; c++)
            hash += getValueInt64(&row, c);
        return hash;
    }
};

/* Check if two Rows are equal */
static bool operator==(const Row& row1, const Row& row2) {
    // Check number of columns.
    if(row1.numCols != row2.numCols)
        return false;
    // Check column types.
    if(row1.types != row2.types)
        for(size_t c = 0; c < row1.numCols; c++)
            if(row1.types[c] != row2.types[c])
                return false;
    // Check stored values.
    for(size_t c = 0; c < row1.numCols; c++)
        if(getValueInt64(&row1, c) != getValueInt64(&row2, c))
            return false;
    // All checks succeeded.
    return true;
}

// ****************************************************************************
// Comparison of two relations
// ****************************************************************************

int matchRelations(Relation* oRes, Relation* eRes) {
    size_t onR = oRes->numRows;
    size_t enR = eRes->numRows;
    size_t onC = oRes->numCols;
    size_t enC = eRes->numCols;

    // The number of columns must match.
    if (onC != 0 && onC != enC) {
        printf("FAIL: Returned %ld columns. Expected %ld columns.\n", onC, enC);
        return EXIT_FAILURE;
    }

    // The number of rows must match.
    if (onR != 0 && onR != enR) {
        printf("FAIL: Returned %ld rows. Expected %ld rows.\n", onR, enR);
        return EXIT_FAILURE;
    }
    
    // The column types must match.
    if (oRes->colTypes != eRes->colTypes)
        for (size_t c = 0; c < onC; c++) {
            DataType oDt = oRes->colTypes[c];
            DataType eDt = eRes->colTypes[c];
            if (oDt != eDt) {
                printf(
                        "FAIL: Column %ld has type %s. Expected %s.\n",
                        c, nameOfDataType(oDt), nameOfDataType(eDt)
                );
                return EXIT_FAILURE;
            }
        }

    // The rows themselves must match (in any order).
    Row* oRow = initRow(oRes);
    std::unordered_set<Row> hs;
    for (size_t r = 0; r < onR; r++) {
        getRow(oRow, oRes, r);
        hs.insert(*oRow);
    }
    freeRow(oRow);
    Row* eRow = initRow(eRes);
    for (size_t r = 0; r < enR; r++) {
        getRow(eRow, eRes, r);
        if (hs.find(*eRow) == hs.end()) {
            printf("FAIL: Result mismatch: expected row %ld not found.\n", r);
            return EXIT_FAILURE;
        }
    }
    freeRow(oRow);

    // All tests succeeded.
    return EXIT_SUCCESS;
}